import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/dialog/save_or_abort_dialog.dart';
import 'package:clean_android_design/src/services/database_service_interface.dart';
import 'package:clean_android_design/src/services/models_interfaces.dart';
import 'package:clean_android_design/src/structure/simple_scaffold_menu_button_widget.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

import 'default_scaffold_widget.dart';
import 'icon_scaffold_menu_button_widget.dart';
import 'scaffold_page_widget.dart';

class DefaultUpdateScaffoldWidget<T extends UpdateableDatabaseItem<T>>
    extends StatefulWidget {
  static DatabaseServiceInterface databaseService;

  final T updateableItem;
  final void Function(BuildContext context, UpdateItemFrame<T> updateItemFrame)
      beforeSave;
  final Future<void> Function(
      BuildContext context, UpdateItemFrame<T> updateItemFrame) beforeSaveAsync;
  final void Function(BuildContext context, DocumentReference itemReference)
      afterInsert;
  final void Function(BuildContext context, UpdateItemFrame<T> updateItemFrame)
      onAbort;
  final void Function(BuildContext context, T item) prepareNewItem;

  final List<Widget> menuButtons;
  final List<Widget> editMenuButtons;
  final bool hasEditButton;
  final bool isDefaultEditable;

  final Widget Function(
      BuildContext context, UpdateItemFrame<T> updateItemFrame) titleBuilder;
  final List<ScaffoldPageWidget<T>> Function(
      BuildContext context,
      UpdateItemFrame<T> updateItemFrame,
      Validator<T> scaffoldValidator) updatePagesBuilder;
  final Widget Function(BuildContext context,
      UpdateItemFrame<T> updateItemFrame, Validator<T> validator) builder;
  final IHelpWidget helpWidget;

  const DefaultUpdateScaffoldWidget({
    Key key,
    this.updateableItem,
    this.menuButtons,
    this.hasEditButton = true,
    this.updatePagesBuilder,
    this.editMenuButtons,
    this.titleBuilder,
    this.beforeSave,
    this.prepareNewItem,
    this.onAbort,
    this.beforeSaveAsync,
    this.afterInsert,
    this.isDefaultEditable = false,
    this.helpWidget,
    this.builder,
  })  : assert(titleBuilder != null),
        assert(beforeSave == null || beforeSaveAsync == null),
        super(key: key);

  @override
  _DefaultUpdateScaffoldWidgetState<T> createState() =>
      _DefaultUpdateScaffoldWidgetState<T>();
}

class _DefaultUpdateScaffoldWidgetState<T extends UpdateableDatabaseItem<T>>
    extends State<DefaultUpdateScaffoldWidget<T>> {
  UpdateItemFrame<T> updateItemFrame;
  bool isNewItem;

  int validatePageIndex;
  Key mainKey;
  Validator<T> validator;

  List<ScaffoldPageWidget<T>> pages;

  @override
  void initState() {
    super.initState();

    if (DefaultUpdateScaffoldWidget.databaseService == null)
      print(
          "[DefaultUpdateScaffoldWidget] Error: The static DatabseService was not set!");

    updateItemFrame = UpdateItemFrame();
    updateItemFrame.originalItem = widget.updateableItem.copy();
    isNewItem = updateItemFrame.originalItem.id == null;
    updateItemFrame.setEditable(isNewItem || widget.isDefaultEditable);

    if (isNewItem)
      widget.prepareNewItem?.call(context, updateItemFrame.editableItem);

    validatePageIndex = null;
    mainKey = Key(Random().nextInt(10000).toString());
  }

  @override
  Widget build(BuildContext context) {
    validator = Validator<T>();
    pages =
        widget.updatePagesBuilder?.call(context, updateItemFrame, validator);

    return DefaultScaffoldWidget(
      key: mainKey,
      title: widget.titleBuilder(context, updateItemFrame),
      showBackArrow: true,
      onWillPop: _onWillPop,
      menuButtons: _getMenuButtons(),
      updatePages: pages,
      child: widget.builder?.call(context, updateItemFrame, validator),
      defaultPageIndex: validatePageIndex ?? 0,
      helpWidget: widget.helpWidget,
    );
  }

  List<Widget> _getMenuButtons() {
    if (!widget.hasEditButton) return widget.menuButtons;

    if (updateItemFrame.isEditable())
      return <Widget>[
        SimpleScaffoldMenuButtonWidget(
          text: Text("Abbrechen"),
          onTap: () => _abortEditing(),
          tooltip: "Bearbeitung abbrechen",
        ),
        SimpleScaffoldMenuButtonWidget(
          text: Text("Speichern"),
          onTap: () => _validate(),
          tooltip: "Änderungen speichern",
        )
      ]..addAll(widget.editMenuButtons ?? widget.menuButtons);
    else
      return <Widget>[
        IconScaffoldMenuButtonWidget(
          icon: Icons.edit,
          text: Text("Bearbeiten"),
          onTap: () async {
            setState(() {
              updateItemFrame.setEditable(true);
            });
          },
          tooltip: "Eigenschaften bearbeiten",
        )
      ]..addAll(widget.menuButtons);
  }

  Future<bool> _onWillPop() async {
    if (updateItemFrame.isEditable()) {
      if (updateItemFrame.editableItem.equals(updateItemFrame.originalItem))
        return true;
      else {
        bool result = false;
        await showCustomDialog(
            context,
            SaveOrAbortDialog(
              onAbort: () {
                result = true;
              },
              onSave: () {
                result = false;
                _validate();
              },
            ));
        return result;
      }
    } else
      return true;
  }

  void _abortEditing() {
    widget.onAbort?.call(context, updateItemFrame);
    setState(() {
      if (isNewItem) {
        Navigator.pop(context);
      } else {
        updateItemFrame.setEditable(false);
      }
    });
  }

  bool _validate() {
    var result = validator.isValid(updateItemFrame.editableItem);
    if (result.isValid) {
      validatePageIndex = null;

      _saveItem();
      return true;
    } else {
      if (pages != null) {
        for (var i = 0; i < pages.length; i++) {
          if (pages[i].key == result.firstInvalidPage) {
            setState(() {
              validatePageIndex = i;
              mainKey = Key(Random().nextInt(10000).toString());
            });
            Future.delayed(Duration(milliseconds: 100), () {
              var result = validator.isValid(updateItemFrame.editableItem);
              if (result.firstInvalidNode != null)
                FocusScope.of(context).requestFocus(result.firstInvalidNode);
            });
          }
        }
      } else {
        setState(() {});
      }
      return false;
    }
  }

  void _saveItem() async {
    if (widget.beforeSave != null) widget.beforeSave(context, updateItemFrame);

    if (widget.beforeSaveAsync != null)
      await widget.beforeSaveAsync(context, updateItemFrame);

    if (isNewItem) {
      var itemRef = await DefaultUpdateScaffoldWidget.databaseService
          .collectionService<T>(updateItemFrame.editableItem)
          .add(updateItemFrame.editableItem);
      widget.afterInsert?.call(context, itemRef);
      Navigator.pop(context);
    } else {
      DefaultUpdateScaffoldWidget.databaseService
          .collectionService<T>(updateItemFrame.editableItem)
          .update(updateItemFrame.editableItem);
      updateItemFrame.setEditable(false);
      updateItemFrame.originalItem = updateItemFrame.editableItem.copy();
    }

    setState(() {});
  }
}

class UpdateItemFrame<T extends UpdateableDatabaseItem<T>> {
  bool _isEditable = false;

  bool isEditable() {
    return _isEditable;
  }

  void setEditable(bool isEditable) {
    _isEditable = isEditable;
    editableItem = originalItem.copy();
  }

  T originalItem;
  T editableItem;
}
